class JOSEError(Exception):
    pass


class JWSError(JOSEError):
    pass


class JWSSignatureError(JWSError):
    pass


class JWSAlgorithmError(JWSError):
    pass


class JWTError(JOSEError):
    pass


class JWTClaimsError(JWTError):
    pass


class JWTSignatureError(JWTError):
    pass


class ExpiredSignatureError(JWTError):
    pass


class JWKError(JOSEError):
    pass


__name__      = 'okta_jwt'
__version__   = '1.3.3'
__author__    = 'Adithya Sampatoor'
__license__   = 'MIT'
__copyright__ = 'Copyright 2018 Adithya Sampatoor'


from .exceptions import JWSSignatureError
from .exceptions import JWTError
from .exceptions import JWTClaimsError
from .exceptions import JWTSignatureError
from .exceptions import ExpiredSignatureError
from .exceptions import JWKError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import okta_jwt
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    long_description = readme.read()


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, '__init__.py'))
    ]


extras_require = {}

setup(
    name="okta_jwt",
    version=okta_jwt.__version__,
    author="Adithya Sampatoor",
    author_email="adithya.sampatoor@gmail.com",
    description="Okta JWT Access Token verifier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/adithyasampatoor/okta_jwt",
    packages=get_packages('okta_jwt'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities'
    ],
    extras_require=extras_require,
    tests_require=['mock', 'ddt'],
    install_requires=['six <2.0', 'requests', 'python-jose']
)



import requests
from jose import jwk, jwt
from requests.auth import HTTPBasicAuth
from jose.utils import base64url_decode
from okta_jwt.utils import verify_exp, verify_aud, check_presence_of, verify_iat, verify_iss, verify_cid



JWKS_CACHE = {}


# Generates Okta Access Token
def generate_token(issuer, client_id, client_secret, username, password, scope='openid'):
    """For generating a token, you need to pass in the Issuer,
    Client ID, Client Secret, Username and Password
    """
    auth = HTTPBasicAuth(client_id, client_secret)

    headers = {
        'Accept':       'application/json',
        'Content-Type': 'application/x-www-form-urlencoded'
    }

    # grant_type is gonna be constant
    payload = {
        "username":   username,
        "password":   password,
        "scope":      scope,
        "grant_type": "password"
    }

    url = "{}/v1/token".format(issuer)

    try:
        response = requests.post(url, data=payload, headers=headers, auth=auth)

        # Consider any status other than 2xx an error
        if not response.status_code // 100 == 2:
            raise Exception(response.text, response.status_code)

        return_value = response.json()

        if 'access_token' not in return_value:
            raise Exception("no access_token in response from /token endpoint", 401)

        print("[Okta::Jwt] Generating Okta Token")
        access_token = return_value['access_token']

        return access_token
    except requests.exceptions.RequestException as e:
        # A serious problem happened, like an SSLError or InvalidURL
        raise Exception("Error: {}".format(str(e)))


# Verifies Claims
def verify_claims(payload, issuer, audience, cid_list):
    """ Validates Issuer, Client IDs, Audience
    Issued At time and Expiration in the Payload
    """
    print("[Okta::Jwt] Verifying Claims")

    verify_iss(payload, issuer)
    verify_cid(payload, cid_list)
    verify_aud(payload, audience)
    verify_exp(payload)
    verify_iat(payload)


# Validates Token
def validate_token(access_token, issuer, audience, client_ids):
    # Client ID's list
    cid_list = []

    if not isinstance(client_ids, list):
        cid_list.append(client_ids)
    else:
        cid_list = client_ids

    check_presence_of(access_token, issuer, audience, cid_list)

    # Decoding Header & Payload from token
    header  = jwt.get_unverified_header(access_token)
    payload = jwt.get_unverified_claims(access_token)

    # Verifying Claims
    verify_claims(payload, issuer, audience, cid_list)

    # Verifying Signature
    jwks_key = fetch_jwk_for(header, payload)
    key      = jwk.construct(jwks_key)
    message, encoded_sig = access_token.rsplit('.', 1)
    decoded_sig = base64url_decode(encoded_sig.encode('utf-8'))

    print("[Okta::Jwt] Validating the Access Token")
    print()

    valid = key.verify(message.encode(), decoded_sig)

    # If the token is valid, it returns the payload 
    if valid == True:
        print('Valid Token')
        return payload
    else:
        raise Exception('Invalid Token')


# Extract public key from metadata's jwks_uri using kid
def fetch_jwk_for(header, payload):
    # Extracting kid from the Header
    if 'kid' in header:
        kid = header['kid']
    else:
        raise ValueError('Token header is missing "kid" value')

    global JWKS_CACHE

    # If there is a matching kid, it wont fetch for kid from the server again
    if JWKS_CACHE:
        if kid in JWKS_CACHE:
            return JWKS_CACHE[kid]

    print("[Okta::Jwt] Fetching public key: kid => " + str(kid))

    # Fetching jwk
    url = fetch_metadata_for(payload)['jwks_uri']

    try:
        jwks_response = requests.get(url)

        # Consider any status other than 2xx an error
        if not jwks_response.status_code // 100 == 2:
            raise Exception(jwks_response.text, jwks_response.status_code)
    except requests.exceptions.RequestException as e:
        # A serious problem happened, like an SSLError or InvalidURL
        raise Exception("Error: {}".format(str(e)))

    jwks = list(filter(lambda x: x['kid'] == kid, jwks_response.json()['keys']))
    if not len(jwks):
        raise Exception("Error: Could not find jwk for kid: {}".format(kid))
    jwk = jwks[0]

    # Adding JWK to the Cache
    JWKS_CACHE[kid] = jwk

    return jwk


def fetch_metadata_for(payload):
    print("[Okta::Jwt] Fetching MetaData")

    # Extracting client_id and issuer from the Payload
    client_id      = payload['cid']
    issuer         = payload['iss']

    # Preparing URL to get the metadata
    url = "{}/.well-known/oauth-authorization-server?client_id={}".format(issuer, client_id)

    try:
        metadata_response = requests.get(url)

        # Consider any status other than 2xx an error
        if not metadata_response.status_code // 100 == 2:
            raise Exception(metadata_response.text, metadata_response.status_code)

        json_obj = metadata_response.json()
        return json_obj

    except requests.exceptions.RequestException as e:
        # A serious problem happened, like an SSLError or InvalidURL
        raise Exception("Error: {}".format(str(e)))



from calendar import timegm
from datetime import datetime, timedelta
from six import string_types
from okta_jwt.exceptions import ExpiredSignatureError, JWTClaimsError



def check_presence_of(access_token, issuer, audience, client_ids):
    if not access_token:
        raise ValueError('Access Token is required')

    if not issuer:
        raise ValueError('Issuer is required')

    if not audience:
        raise ValueError('Audience is required')

    if '' in client_ids:
        raise ValueError('Client ID is required')


def verify_iss(payload, issuer):
    if payload['iss'] != issuer:
        raise JWTClaimsError('Invalid Issuer')


def verify_cid(payload, cid_list):
    if not payload['cid'] in cid_list:
        raise JWTClaimsError('Invalid Client')


def verify_exp(payload, leeway=0):
    """Validates that the 'exp' claim is valid.
    The "exp" (expiration time) claim identifies the expiration time on
    or after which the JWT MUST NOT be accepted for processing.  The
    processing of the "exp" claim requires that the current date/time
    MUST be before the expiration date/time listed in the "exp" claim.
    Implementers MAY provide for some small leeway, usually no more than
    a few minutes, to account for clock skew.
    Args:
        payload (dict): The payload dictionary to validate.
        leeway (int): The number of seconds of skew that is allowed.
    """
    if 'exp' not in payload:
        return

    try:
        exp = int(payload['exp'])
    except ValueError:
        raise JWTClaimsError('Expiration Time payload (exp) must be an integer.')

    now = timegm(datetime.utcnow().utctimetuple())

    if exp < (now - leeway):
        raise ExpiredSignatureError('Token is expired.')


def verify_aud(payload, audience=None):
    """Validates that the 'aud' claim is valid.
    The "aud" (audience) claim identifies the recipients that the JWT is
    intended for. Each principal intended to process the JWT MUST
    identify itself with a value in the audience claim.  If the principal
    processing the claim does not identify itself with a value in the
    "aud" claim when this claim is present, then the JWT MUST be
    rejected.  In the general case, the "aud" value is an array of case-
    sensitive strings, each containing a StringOrURI value.
    """
    if 'aud' not in payload:
        return

    audience_claims = payload['aud']

    if isinstance(audience_claims, string_types):
        audience_claims = [audience_claims]
    if not isinstance(audience_claims, list):
        raise JWTClaimsError('Invalid claim format in token')
    if any(not isinstance(c, string_types) for c in audience_claims):
        raise JWTClaimsError('Invalid claim format in token')
    if audience not in audience_claims:
        raise JWTClaimsError('Invalid Audience')


def verify_iat(payload, leeway=300):
    """The iat value indicates what time the token was "issued at". 
    We verify that this claim is valid by checking that the token was 
    not issued in the future, with some leeway for clock skew.
    """
    time_now_with_leeway = datetime.utcnow() + timedelta(seconds=leeway)
    acceptable_iat = timegm((time_now_with_leeway).timetuple())

    if 'iat' in payload and payload['iat'] > acceptable_iat:
        raise JWTClaimsError('Invalid Issued At(iat) Time')
